--
-- Model: Cooperation (greedy vs cooperative cows)
--
-- Author: Copyright 1997 Uri Wilensky (http://ccl.northwestern.edu/netlogo/models/Cooperation)
--         TerraME version adapted and implemented by Merret Buurman
--         (as an assignment in the class "Environmental Modelling" taught by Prof. Tiago Carneiro)
--
-- References:
-- * Wilensky, U. (1997). NetLogo Cooperation model. http://ccl.northwestern.edu/netlogo/models/Cooperation. Center for Connected Learning and Computer-Based Modeling, Northwestern University, Evanston, IL.
-- * Wilensky, U. (1999). NetLogo. http://ccl.northwestern.edu/netlogo/. Center for Connected Learning and Computer-Based Modeling, Northwestern University, Evanston, IL.
--
-- TerraME model with the following structure:
--              1) Define space
--				2) Define behavior
--				3) Define time
--				4) Run simulations 
--


options = {
	-- You can either define the options in this table...
	initial_cows            = 10, -- How many cows will be there initially?
	cellspacesize           = 20,   -- What size will the space have?
}

-- Or you can define them here:
options.stride_length = 1
options.printbirthdeath = false

-- For how many time steps should the model run?
totaltimesteps = 500


run_model = function(useroptions, totaltimesteps)

	----------------------
	-- Default options: --
	----------------------

	options = {} -- Note: Whenever the user could specify "false", but the default is something true, need to put an if clause here, instead of "or"!

	-- Initial settings:
	options.initial_cows            = useroptions.initial_cows 				or 10    -- How many cows will be there initially?
	options.cellspacesize           = useroptions.cellspacesize 			or 20    -- What size will the space have?
	options.cooperative_probability = useroptions.cooperative_probability 	or 0.5   -- What fraction of cows will be greedy or cooperation, in the beginning?
	options.number_greedy_beginning = useroptions.number_greedy_beginning 	or false -- How many cows will be greedy in the beginning? (Overrides previous parameter)

	-- Cow options:
	options.grass_energy           = useroptions.grass_energy 			or 50   -- How much energy it gains by eating one grass unit
	options.metabolism             = useroptions.metabolism 			or 40   -- How much energy it loses per timestep
	options.reproduction_threshold = useroptions.reproduction_threshold or 300  -- At which energy level will it reproduce?
	options.reproduction_cost      = useroptions.reproduction_cost 		or 50   -- How much energy will the cow lose when it reproduces?
	options.stride_length          = useroptions.stride_length 			or 1    -- How far does a cow walk in each time step?

	-- Possible advantages of cooperative cows:
	options.energylevel_for_advantages            = useroptions.energylevel_for_advantages 				or 200
	options.advantage_lose_less_in_reproduction   = useroptions.advantage_lose_less_in_reproduction 	or false
	options.advantage_lose_less_in_metabolism     = useroptions.advantage_lose_less_in_metabolism 		or false
	options.advantage_can_eat_more                = useroptions.advantage_can_eat_more 					or false
	options.advantage_gets_more_energy_from_food  = useroptions.advantage_gets_more_energy_from_food 	or false
	if (useroptions.advantage_more_resistant_to_parasites==false) then options.advantage_more_resistant_to_parasites = false else options.advantage_more_resistant_to_parasites = true end

	-- Grass options:
	options.low_high_threshold     = useroptions.low_high_threshold or 5
	options.max_grass_height       = useroptions.max_grass_height   or 10
	options.high_growth_chance     = useroptions.high_growth_chance or 0.9
	options.low_growth_chance      = useroptions.low_growth_chance  or 0.3

	-- Parasites:
	options.parasites 			= useroptions.parasites 			or false 	-- Should parasites be simulated?
	options.parasites_start 	= useroptions.parasites_start		or 1 		-- When should the parasites enter the simulation?
	options.parasites_number 	= useroptions.parasites_number		or 1 		-- How many cows should initially be infected?
	options.parasitenenergyloss = useroptions.parasitenenergyloss 	or 0.15 	-- How much energy (fraction of its energy, from 0 to 1) does a cow lose per time step?
	options.healthycow_level 	= useroptions.healthycow_level		or 350 		-- At what level of energy should a cow get rid of the parasites? Should be larger than reproduction level!

	-- Simulation options
	options.wait 					= useroptions.wait 					or false 	-- How much time the simulation should wait per timestep (if no waiting is desired, put "false")
	if (useroptions.printbirthdeath==false) then options.printbirthdeath = false else options.printbirthdeath = true end -- Should information about births and deaths be printed to screen?
	options.randomseed 				= useroptions.randomseed			or false    -- In case you want to fix the random seed for reproduceable results 	
	options.store_resultfiles		= useroptions.store_resultfiles 	or false    -- Should result files be stored?	
	options.path_for_resultfiles 	= useroptions.path_for_resultfiles	or "" 		-- Where should the result .txt be saved?

	-- Which observers?
	if (useroptions.observers == false) then options.observers = useroptions.observers 				 		else options.observers = true end -- Should observers be displaxed at all?
	if (useroptions.killobservers == false) then options.killobservers = useroptions.killobservers 			else options.killobservers = true end -- Should observers be killed at the end (useful when running the model many times, in Monte Carlo simulation)
	if (useroptions.observe_map == false) then options.observe_map = useroptions.observe_map 				else options.observe_map = true end  --   Map of grass and agents
	if (useroptions.observe_cownumbers == false) then options.observe_cownumbers = useroptions.observe_cownumbers 		else options.observe_cownumbers = true end  --   Chart of population numbers
	if (useroptions.observe_grassenergy == false) then options.observe_grassenery = useroptions.observe_grassenery 		else options.observe_grassenery = true end  --   Chart of overall grass amount
	if (useroptions.observe_cowenergy == false) then options.observe_cowenergy = useroptions.observe_cowenergy 			else options.observe_cowenergy = true end  --   Chart of cows' energy
	if (useroptions.observe_grass_and_cows == false) then options.observe_grass_and_cows = useroptions.observe_grass_and_cows 	else options.observe_grass_and_cows = true end  --   Chart of population number and grass amount (to observe prey and predator patterns. Here, grass amount is adapted to fit into the chart.)
	if (useroptions.observe_parasites == false) then options.observe_parasites = useroptions.observe_parasites					else options.observe_parasites = true end --   Chart of percentage of parasite infection
	if (useroptions.observe_averagecows == false) then options.observe_averagecows = useroptions.observe_averagecows 			else options.observe_averagecows = true end  --   Chart of average cows' energy level

	print("Running the model with the following options:")
	for key,val in pairs(options) do
		print("   "..key,val)
	end
	print()

	---------------------------
	-- Some general settings --
	---------------------------

	random = Random{seed = options.randomseed}

	local wait = function(seconds)
		local start = os.clock()
		repeat until os.clock() > start + seconds
	end

	firsttime = true -- for writing header to result file

	-----------
	-- Space --
	-----------

	local pasture = CellularSpace{
		xdim = options.cellspacesize,

		-- Statistics, not needed for functioning:
		zero = 0,
		number_of_cows = 0,
		number_of_cows_coop = 0,
		number_of_cows_greedy = 0,
		nomorecoop = false,
		nomoregreedy = false,

		amount_of_grass = 0,
		amount_of_grass_lower = 0,

		amount_of_cowenergy = 0,
		amount_of_cowenergy_coop = 0,
		amount_of_cowenergy_greedy = 0,

		average_energy = 0,
		average_energy_coop = 0,
		average_energy_greedy = 0,

		number_of_infected = 0,
		number_of_infected_greedy = 0,
		number_of_infected_coop = 0,
		percentage_infected = 0,
		percentage_infected_coop = 0,
		percentage_infected_greedy = 0,
		num_COOP_NOPARA = 0,
		num_COOP_PARA = 0,
		num_GREEDY_NOPARA = 0,
		num_GREEDY_PARA = 0
	}

	pasture:createNeighborhood{
		strategy = "moore", -- 8 neighbours!
		name = "nh"
	}


	------------------------
	-- Behaviour of space --
	------------------------

	local grassgrowth_cell = function(previousgrassheight)
		local newgrassheight = previousgrassheight

		-- High grass
		if (previousgrassheight >= options.low_high_threshold) then
			if (options.high_growth_chance >= random:number()) then
				newgrassheight = previousgrassheight+1
			end

		-- Low grass
		else
			if (options.low_growth_chance >= random:number()) then
				newgrassheight = previousgrassheight+1
			end
		end

		-- Too high?
		if (newgrassheight > options.max_grass_height) then
			newgrassheight = options.max_grass_height
		end

		return(newgrassheight)
	end

	local growGrass_inwhole = function(cs)
		forEachCell(cs, function(cell)
			cell.grassheight = grassgrowth_cell(cell.grassheight)
		end)
	end

	-----------------------
	-- Behaviour of cows --
	-----------------------

	-- Constants:
	-- Behaviour
	COOPERATIVE = 1
	GREEDY = 2
	-- with parasites (for being able to distinguish four types in a legend):
	COOP_PARA = 1
	COOP_NOPARA = 2
	GREEDY_PARA = 3
	GREEDY_NOPARA = 4
	-- Strings for printing etc.
	behav = {}
	behav[COOPERATIVE] = "cooperative"
	behav[GREEDY] = "greedy"

	local cow = Agent{

		-- behaviour + health:
		behaviour = 0,
		legendstate = 0,
		parasite = false,
		energylevel = options.metabolism * 4,
		age = 0, -- not really needed
		
		-- location:
		precisexcoord = nil,
		preciseycoord = nil,

		-- reference to observer so I can kill it when cow dies!
		observer = nil,

		-- Behaviour
		eat = function(agent, cell)

			--local prev = cell.grassheight

			-- Cooperative agent:
			if (agent.behaviour == COOPERATIVE) then
				if (cell.grassheight > options.low_high_threshold) then

					local eaten = 0
					local gain_energy_per_food_unit = 0

					-- How much grass does it eat?
					if (options.advantage_can_eat_more) and (agent.energylevel >= options.energylevel_for_advantages) then
						eaten = 2
						print("Cooperative cow "..agent.id.." is using its advantage that it can eat more!")
					else
						eaten = 1
					end

					-- How much energy does it get?
					if (options.advantage_gets_more_energy_from_food) and (agent.energylevel >= options.energylevel_for_advantages) then
						gain_energy_per_food_unit = 1.5
						print("Cooperative cow "..agent.id.." is using its advantage that it gains more energy from the food!")
					else
						gain_energy_per_food_unit = 1
					end

					cell.grassheight = cell.grassheight - eaten
					agent.energylevel = agent.energylevel + gain_energy_per_food_unit*eaten*options.grass_energy
					--print("Eaten: "..eaten..", energy gained: "..gain_energy_per_food_unit*eaten..", energylevel: ".. agent.energylevel)
				end

			-- Egoist agent:
			elseif (agent.behaviour == GREEDY) then
				if (cell.grassheight > 0) then
					cell.grassheight = cell.grassheight -1
					agent.energylevel = agent.energylevel + options.grass_energy
				end
			end

			--print("Cell "..cell.x..","..cell.y.." was grazed from "..prev.." to "..cell.grassheight)
		end,
		metabolism = function(agent)

			if (agent.behaviour == COOPERATIVE) and (options.advantage_lose_less_in_metabolism) and (agent.energylevel >= options.energylevel_for_advantages) then
				agent.energylevel = agent.energylevel - 0.5*options.metabolism
				print("Cooperative cow "..agent.id.." is using its metabolism advantage!")
			else
				agent.energylevel = agent.energylevel - options.metabolism
			end
			if (agent.parasite) then
				agent.energylevel = (1-options.parasitenenergyloss)*agent.energylevel
			end
		end,
		reproduction_if_applicable = function(agent)
			if (agent.energylevel >= options.reproduction_threshold) then

				-- Parent reproduced and loses energy
				calf = agent:reproduce()
				if (agent.behaviour == COOPERATIVE) and (options.advantage_lose_less_in_reproduction) and (agent.energylevel >= options.energylevel_for_advantages) then
					agent.energylevel = agent.energylevel - 0.5*options.reproduction_cost
					print("Cooperative cow "..agent.id.." is using its reproductive advantage!")
				else
					agent.energylevel = agent.energylevel - options.reproduction_cost
				end

				-- New cow is initialized:
				if (agent.behaviour == COOPERATIVE) then
					calf.behaviour = COOPERATIVE
					calf.legendstate = COOP_NOPARA
				else
					calf.behaviour = GREEDY
					calf.legendstate = GREEDY_NOPARA
				end

				-- Print!
				if (options.printbirthdeath) then print(behav[agent.behaviour].." cow "..agent.id.." reproduced ("..calf.id..": "..behav[calf.behaviour]..").") end
				
				-- Add observer to new agent:
				if (options.observe_map) then
					calf.observer = Observer{
						subject = calf,
						type = "map",
						observer = map,
						attributes = {"legendstate"},
						legends = {cowlegend_withparasites}
					}
				end
			end
		end,
		death_if_applicable = function(agent)
			if (agent.energylevel <= 0) then
				if (options.printbirthdeath) then print(behav[agent.behaviour].." cow "..agent.id.." died at age "..agent.age..".") end
				if (agent.observer) then
					agent.observer:kill()
				end
				agent:die()
			end
		end,
		moveon = function(agent)

			local debugthis = false

			if debugthis then
				print("")
				print("Move "..agent.id.."?")
				print("Previous location:",agent:getCell().x,   agent:getCell().y)
				print("         precise: ",agent.precisexcoord, agent.preciseycoord)
			end

			-- Make a move of length options.stride_length in any direction
			local movex = random:number(-options.stride_length, options.stride_length) -- going to both x directions
			local movey = math.sqrt(math.pow(options.stride_length,2)+math.pow(movex,2))
			if (random:number() < 0.5) then movey = -1 * movey end -- to avoid just always going in one y direction!
			agent.precisexcoord = agent.precisexcoord + movex
			agent.preciseycoord = agent.preciseycoord + movey
			if debugthis then print("     new precise: ",agent.precisexcoord, agent.preciseycoord) end

			-- Now round the new coordinates to integers
			local xcoord = round(agent.precisexcoord)
			local ycoord = round(agent.preciseycoord)
			if debugthis then print("  after rounding: ",xcoord, ycoord) end
			
			-- Check if the move lead out of the current cell:
			if (agent:getCell().x ~= xcoord) or (agent:getCell().y ~= ycoord) then
				if debugthis then
					print("Need to move!")
					prevx = agent:getCell().x
					prevy = agent:getCell().y
				end

				--[[ If the move leads out of the cell space, go back in on the other side!
				if (xcoord+1 > pasture:getXDim()) then  -- +1 because when xdim is 10, the highest coordinate is 9, as they start with 0!
					xcoord = xcoord - pasture:getXDim()
					agent.precisexcoord = agent.precisexcoord - pasture:getXDim()
				end
				if (ycoord+1 > pasture:getYDim()) then
					ycoord = ycoord - pasture:getYDim()
					agent.preciseycoord = agent.preciseycoord - pasture:getYDim()
				end
				if (xcoord < 0) then
					xcoord = xcoord + pasture:getXDim()
					agent.precisexcoord = agent.precisexcoord + pasture:getXDim()
				end
				if (ycoord < 0) then
					ycoord = ycoord + pasture:getYDim()
					agent.preciseycoord = agent.preciseycoord + pasture:getYDim()
				end
				--]]
				--If the move leads out of the cell space, go back in!
				if (xcoord+1 > pasture:getXDim()) then  -- +1 because when xdim is 10, the highest coordinate is 9, as they start with 0!
					xcoord = pasture:getXDim() - 1
					agent.precisexcoord = xcoord - random:number(0,0.8)
				end
				if (ycoord+1 > pasture:getYDim()) then
					ycoord = pasture:getYDim() - 1
					agent.preciseycoord = ycoord - random:number(0,0.8)
				end
				if (xcoord < 0) then
					xcoord = 0
					agent.precisexcoord = random:number(0,0.8)
				end
				if (ycoord < 0) then
					ycoord = 0
					agent.preciseycoord = random:number(0,0.8)
				end

				-- Now carry out the move:
				local coord = Coord{x = xcoord, y = ycoord} 
				agent:move(pasture:getCell(coord))
				if debugthis then print("Moved from "..prevx..","..prevy.." to "..xcoord..","..ycoord.."!") end
			end

			--[[local shouldmove = function()
				local randomnumber = random:number()
				if (randomnumber < options.stride_length) then
					return(true)
				else
					return(false)
				end
			end

			if (shouldmove()) then
				--agent:randomWalk("nh")
				local nextcell = cell:getNeighborhood("nh"):sample()
				agent:move(nextcell)
			end
			--]]

			--print("Cow "..agent.id.." moved to cell ",nextcell.x, nextcell.y)
		end,
		cure_if_applicable = function(agent)
			if (agent.parasite) and (agent.energylevel > options.healthycow_level) then
				agent.parasite = false
				print(agent.id.." is cured from parasites!")
			end
		end,


		execute = function(self)

			-- First time:
			if (self.precisexcoord == nil) then
				self.precisexcoord = self:getCell().x
				self.preciseycoord = self:getCell().y
			end

			--print("Cow "..self.id.." is doing stuff!")
			self.age = self.age + 1
			self:moveon(self:getCell())
			self:eat(self:getCell())
			self:metabolism()
			self:reproduction_if_applicable()
			self:cure_if_applicable()
			self:death_if_applicable()
		end,

		init = function(self)

			-- This function is called in the beginning, when the herd is created, and for every reproduction.
			-- The default values of the attributes were already set.
			--print("Age of new cow "..self.id..": "..self.age..", energy level: "..self.energylevel..", behaviour: "..self.behaviour)

			-- I do not set the behaviour here, as I want it to be inherited from the parents.

			-- Can not call this yet! Had to move it to execute()
			--self.precisexcoord = self:getCell().x
			--self.preciseycoord = self:getCell().y
		end
	}

	---------------
	-- Parasites --
	---------------
	herd = Society {
	    instance = cow,
	    quantity = options.initial_cows
	}

	local setparasites = function()

		if (options.parasites) then
			local infectedcows = 0
			while (infectedcows < options.parasites_number) do
				candidate = herd:sample()
				if (candidate.parasite == false) then
					candidate.parasite = true
					infectedcows = infectedcows + 1
					if (candidate.behaviour == COOPERATIVE) then
						candidate.legendstate = COOP_PARA
					else
						candidate.legendstate = GREEDY_PARA
					end
				end
			end
		end
	end
	local parasites = function()

		if (options.parasites) then

			-- Infection of other cows that are in the same cell:
			forEachAgent(herd, function(cow)
				local neighbourcows = cow:getCell():getAgents()
				if (cow.parasite) and (#neighbourcows>1) then
					local counter=0
					--print("There is "..#neighbourcows.." in the same cell!")
					for key,agent in pairs(neighbourcows) do
						if (agent.parasite == false) then
							local infectionprobability = 1 - (agent.energylevel / (options.healthycow_level))
							if (agent.behaviour == COOPERATIVE) and (options.advantage_more_resistant_to_parasites) and (agent.energylevel >= options.energylevel_for_advantages) then
								local oldrisk = infectionprobability
								infectionprobability = 1 - (1.5*agent.energylevel / (options.healthycow_level))
								print("Cooperative cow "..cow.id.." is using its advantage that it has a lower infection risk ("..infectionprobability.." instead of "..oldrisk..")!")
							end
							local ran = random:number()
							--print("Infection probability: ".. infectionprobability, "ran="..ran)
							if (ran < infectionprobability) then
								agent.parasite = true
								if (agent.behaviour == COOPERATIVE) then
									agent.legendstate = COOP_PARA
								else
									agent.legendstate = GREEDY_PARA
								end
								counter = counter+1
							end
						end
					end
					--print("Infected "..counter.." other cows!")
				end
			end)

		end
	end

	----------------
	-- Statistics --
	----------------

	local globalGrassStats = function(cs)
		cs.amount_of_grass = 0

		forEachCell(cs, function(cell)
			cs.amount_of_grass = cs.amount_of_grass + cell.grassheight
		end)

		cs.amount_of_grass_lower = 0.25 * cs.amount_of_grass
	 end

	local globalEnergyStats = function(cs, soc)

		-- cow energy
		cs.amount_of_cowenergy        = 0
		cs.amount_of_cowenergy_coop   = 0
		cs.amount_of_cowenergy_greedy = 0

		-- average
		cs.average_energy        = 0
		cs.average_energy_greedy = 0
		cs.average_energy_coop   = 0

		forEachAgent(soc, function(ag)
			cs.amount_of_cowenergy = cs.amount_of_cowenergy + ag.energylevel
			if (ag.behaviour == COOPERATIVE) then
				cs.amount_of_cowenergy_coop = cs.amount_of_cowenergy_coop + ag.energylevel
			elseif (ag.behaviour == GREEDY) then
				cs.amount_of_cowenergy_greedy = cs.amount_of_cowenergy_greedy + ag.energylevel
			end
		end)

		-- Compute averages (call it after having done the stats for cow numbers!!)
		cs.average_energy = cs.amount_of_cowenergy / cs.number_of_cows
		cs.average_energy_coop = cs.amount_of_cowenergy_coop / cs.number_of_cows_coop
		cs.average_energy_greedy = cs.amount_of_cowenergy_greedy / cs.number_of_cows_greedy
		if (cs.number_of_cows        == 0) then cs.average_energy = 0 end
		if (cs.number_of_cows_coop   == 0) then cs.average_energy_coop = 0 end
		if (cs.number_of_cows_greedy == 0) then cs.average_energy_greedy = 0 end
	end

	local globalCowStats = function(cs, soc, timestep)

		-- cow numbers
		cs.number_of_cows        = 0
		cs.number_of_cows_coop   = 0
		cs.number_of_cows_greedy = 0

		-- infected cow numbers
		cs.number_of_infected         = 0
		cs.number_of_infected_coop    = 0
		cs.number_of_infected_greedy  = 0
		cs.percentage_infected        = 0
		cs.percentage_infected_coop   = 0
		cs.percentage_infected_greedy = 0

		-- by behaviour and infection:
		cs.num_COOP_NOPARA   = 0
		cs.num_COOP_PARA     = 0
		cs.num_GREEDY_NOPARA = 0
		cs.num_GREEDY_PARA   = 0

		-- Count cow numbers
		forEachAgent(soc, function(ag)

			-- Disregarding behaviour:
			cs.number_of_cows = cs.number_of_cows + 1
			if (ag.parasite) then cs.number_of_infected = cs.number_of_infected + 1 end

			-- Distinguished by behaviour:
			if (ag.behaviour == COOPERATIVE) then
				cs.number_of_cows_coop = cs.number_of_cows_coop + 1
				cs.amount_of_cowenergy_coop = cs.amount_of_cowenergy_coop + ag.energylevel
				if (ag.parasite) then cs.number_of_infected_coop = cs.number_of_infected_coop + 1 end

			elseif (ag.behaviour == GREEDY) then
				cs.number_of_cows_greedy = cs.number_of_cows_greedy + 1
				cs.amount_of_cowenergy_greedy = cs.amount_of_cowenergy_greedy + ag.energylevel
				if (ag.parasite) then cs.number_of_infected_greedy = cs.number_of_infected_greedy + 1 end
			end

			-- Distinguished by behaviour and infection state:
			if (ag.legendstate==COOP_NOPARA) then
				cs.num_COOP_NOPARA = cs.num_COOP_NOPARA + 1
			elseif (ag.legendstate==COOP_PARA) then
				cs.num_COOP_PARA = cs.num_COOP_PARA + 1
			elseif (ag.legendstate==GREEDY_NOPARA) then
				cs.num_GREEDY_NOPARA = cs.num_GREEDY_NOPARA + 1
			elseif (ag.legendstate==GREEDY_PARA) then
				cs.num_GREEDY_PARA = cs.num_GREEDY_PARA + 1
			else
				error("This legend state does not exist!")
			end
		end)


		-- Compute percentages
		cs.percentage_infected                                                   = cs.number_of_infected        / cs.number_of_cows
		if (cs.number_of_infected_coop > 0)   then cs.percentage_infected_coop   = cs.number_of_infected_coop   / cs.number_of_cows_coop end
		if (cs.number_of_infected_greedy > 0) then cs.percentage_infected_greedy = cs.number_of_infected_greedy / cs.number_of_cows_greedy end


		-- Check if values are ok?
		if (cs.number_of_cows ~= (cs.number_of_cows_greedy + cs.number_of_cows_coop)) then
			error("Cows="..cs.number_of_cows,"\n   Greedy="..cs.number_of_cows_greedy.."\n     Coop="..cs.number_of_cows_coop)
		end
		if (cs.number_of_infected ~= (cs.number_of_infected_greedy + cs.number_of_infected_coop)) then
			error("Infected cows="..cs.number_of_infected,"\n   Greedy="..cs.number_of_infected_greedy.."\n     Coop="..cs.number_of_infected_coop)
		end
		if (cs.number_of_infected ~= (cs.number_of_infected_greedy + cs.number_of_infected_coop)) then
			error("Infected cows="..cs.number_of_infected,"\n   Greedy="..cs.number_of_infected_greedy.."\n     Coop="..cs.number_of_infected_coop)
		end
		if ((cs.num_GREEDY_PARA+cs.num_GREEDY_NOPARA+cs.num_COOP_PARA+cs.num_COOP_NOPARA) ~= (cs.number_of_cows_coop + cs.number_of_cows_greedy)) then
			print("Greedy cows:     ",cs.number_of_cows_greedy,"= "..cs.num_GREEDY_NOPARA.." + "..cs.num_GREEDY_PARA)
			print("Cooperative cows:",cs.number_of_cows_coop,  "= "..cs.num_COOP_NOPARA  .." + "..cs.num_COOP_PARA)
			error("Wrong number of cows!")
		end

		if (cs.number_of_cows_coop   == 0) and (cs.nomorecoop==false) then
			print(" ********************** NO MORE COOPERATIVE COWS AT TIME "..timestep.."\n  "..cs.amount_of_cowenergy_coop)
			cs.nomorecoop = true
			if (timestep > 2) then wait(3) end
		end
		if (cs.number_of_cows_greedy == 0) and (cs.nomoregreedy==false) then
			print(" ********************** NO MORE GREEDY COWS AT TIME "..timestep.."\n  "..cs.amount_of_cowenergy_greedy)
			cs.nomoregreedy = true
			if (timestep > 2) then wait(3) end
		end
	end

	globalWriteStats = function(cs, herd, timestep)

		if (options.store_resultfiles) then

			-- Make overview string
			local string_sumup = timestep..";"..cs.amount_of_grass..";"..cs.num_COOP_NOPARA..";"..cs.num_COOP_PARA..";"..cs.num_GREEDY_NOPARA..";"..cs.num_GREEDY_PARA
			local file_sumup = io.open(options.path_for_resultfiles.."results.txt", "a+")
			if (firsttime) then
				local string_sumup = "timestep;totalgrass;num_COOP_NOPARA;num_COOP_PARA;num_GREEDY_NOPARA;num_GREEDY_PARA"
				file_sumup:write("\n"..string_sumup)
			end
			file_sumup:write("\n"..string_sumup)
			file_sumup:close()

			-- Make individuals' string
			local energy_coop_nopara = "xyz"
			local energy_coop_para = "xyz"
			local energy_greedy_nopara = "xyz"
			local energy_greedy_para = "xyz"
			forEachAgent(herd, function(agent)
				if (agent.legendstate==COOP_NOPARA) then
					energy_coop_nopara = energy_coop_nopara..","..agent.energylevel
				elseif (agent.legendstate==COOP_PARA) then
					energy_coop_para = energy_coop_para..","..agent.energylevel
				elseif (agent.legendstate==GREEDY_NOPARA) then
					energy_greedy_nopara = energy_greedy_nopara..","..agent.energylevel
				elseif (agent.legendstate==GREEDY_PARA) then
					energy_greedy_para = energy_greedy_para..","..agent.energylevel
				else error("Does not exist") end
			end)
			local energy_cells = "xyz"
			forEachCell(cs, function(cell)
				energy_cells = energy_cells..","..cell.grassheight
			end)
			energy_cells = string.gsub(energy_cells, "xyz,", "")
			energy_coop_nopara = string.gsub(energy_coop_nopara, "xyz,", "")
			energy_coop_para = string.gsub(energy_coop_para, "xyz,", "")
			energy_greedy_nopara = string.gsub(energy_greedy_nopara, "xyz,", "")
			energy_greedy_para = string.gsub(energy_greedy_para, "xyz,", "")
			local string_indiv = timestep..";"..energy_cells..";"..energy_coop_nopara..";"..energy_coop_para..";"..energy_greedy_nopara..";"..energy_greedy_para
			local file_indiv = io.open(options.path_for_resultfiles.."individualresults.txt", "a+")
			if (firsttime) then
				local string_indiv = "timestep;all_cells_grassenergy;energy_coop_nopara;energy_coop_para;energy_greedy_nopara;energy_greedy_para"
				file_indiv:write("\n"..string_indiv)
			end
			file_indiv:write("\n"..string_indiv)
			file_indiv:close()

			firsttime = false

		end

	end


	----------
	-- Time --
	----------

	local t = Timer{
		Event{action = function(e)
			timestep = e:getTime()

			print("")
			print("Timestep",timestep)

			pasture:synchronize()
			growGrass_inwhole(pasture)
			herd:execute()
			if (options.parasites_start) and (timestep == options.parasites_start) then setparasites() end
			if (options.parasites_start) and (timestep >= options.parasites_start) then parasites() end

			-- Observing:
			globalGrassStats(pasture)
			globalCowStats(pasture, herd, timestep)
			globalEnergyStats(pasture, herd)
			globalWriteStats(pasture, herd, timestep)
			pasture:notify(timestep)

			if (options.wait) then wait(options.wait) end

		end}
	}

	-----------------
	-- Environment --
	-----------------

	env = Environment {pasture, herd, t}
	env:createPlacement {
		strategy = "random",
		--max = 1
	}

	----------------------------
	-- Set initial conditions --
	----------------------------

	-- Initialize cell attributes with zeros, so that the attribute exists (for the observer):
	print()
	print("Putting grass into the pasture...")
	forEachCell(pasture, function(cell)
		cell.grassheight = options.max_grass_height
	end)

	--pasture:synchronize() -- past not used!
	pasture:notify(timestep)

	if (options.number_greedy_beginning) then
		print("Giving specified behaviour to first batch of cows:")
		local num_greedy = 0
		forEachAgent(herd, function(agent)
			if (num_greedy < options.number_greedy_beginning) then
				agent.behaviour = GREEDY
				agent.legendstate = GREEDY_NOPARA
				num_greedy = num_greedy + 1
			else
				agent.behaviour = COOPERATIVE
				agent.legendstate = COOP_NOPARA
			end
		end)
		--[[local num_greedy = 0
		while (num_greedy < options.number_greedy_beginning) do --number_greedy_beginning
			local nextcow = herd:sample()
			if (nextcow.behaviour == COOPERATIVE) then
				nextcow.behaviour = GREEDY
				nextcow.legendstate = GREEDY_NOPARA
				num_greedy = num_greedy + 1
			end
		end--]]
	else
		print("Giving random behaviour to first batch of cows:")
		forEachAgent(herd, function(agent)
			if (random:number() < options.cooperative_probability) then
				agent.behaviour = COOPERATIVE
				agent.legendstate = COOP_NOPARA
			else
				agent.behaviour = GREEDY
				agent.legendstate = GREEDY_NOPARA
			end
		end)
	end
	-- Printing:
	forEachAgent(herd, function(agent)
		if (agent.behaviour == COOPERATIVE) then
			print("  Cow "..agent.id.." is cooperative.")
		else
			print("  Cow "..agent.id.." is egoist.")
		end
	end)

	---------------
	-- Observers --
	---------------

	observers = {}

	-- Legends --

	grasslegend = Legend {
	    grouping = "equalsteps",
	    slices = 50,
	    colorBar = {
	    	--{value = 0, color = "red"},
	    	--{value = options.low_high_threshold, color = "white"},
	        {value = 0, color = "white"},
	        {value = options.max_grass_height, color = "darkGreen"}
	    }
	}
	--table.insert(observers, grasslegend)
	cowlegend = Legend {
	    grouping = "uniquevalue",
		symbol = "N",
		font = "Animals",
		fontSize = 12,
	    colorBar = {
	        {value = COOPERATIVE, color = "blue"},
	        {value = GREEDY, color = "red"}
	    }
	}
	--table.insert(observers, cowlegend)
	cowlegend_withparasites = Legend {
	    grouping = "uniquevalue",
		symbol = "N",
		font = "Animals",
		fontSize = 12,
	    colorBar = {
	    	{value = 0, color = "white"},
	        {value = COOP_NOPARA, color = "darkBlue"},
	        {value = GREEDY_NOPARA, color = "darkMagenta"},
	        --{value = GREEDY_NOPARA, color = "red"},
	        {value = COOP_PARA, color = "blue"},
	        {value = GREEDY_PARA, color = "lightMagenta"}
	    }
	}
	--table.insert(observers, cowlegend_withparasites)

	-- Chart legend
	legend_chart_grass = Legend{
		--grouping = "uniquevalue",
		colorBar = {
			{value = 0, color = "darkGreen"}
		},
		symbol = "none",
		width = 1,
		style = "lines"
	}
	legend_chart_egoist = Legend{
		--grouping = "uniquevalue",
		colorBar = {
			{value = 0, color = "red"}
		},
		symbol = -1,
		width = 1,
		style = "lines"
	}
	legend_chart_cooperative = Legend{
		--grouping = "uniquevalue",
		colorBar = {
			{value = 0, color = "blue"}
		},
		symbol = -1,
		width = 1,
		style = "lines"
	}
	legend_chart_zero = Legend{
		colorBar = {
			{value = 0, color = "lightGray"}
		},
		symbol = -1,
		width = 1,
		style = "lines"
	}
	legend_chart_totalcows = Legend{
		colorBar = {
			{value = 0, color = "black"}
		},
		--symbol = "none",
		symbol = -1,
		width = 1,
		style = "lines"
	}
	--[[
	table.insert(observers, legend_chart_grass)
	table.insert(observers, legend_chart_totalcows)
	table.insert(observers, legend_chart_zero)
	table.insert(observers, legend_chart_cooperative)
	table.insert(observers, legend_chart_egoist)
	--]]

	-- Map --

	if (options.observe_map) then
		map = Observer{
			subject = pasture,
			attributes = {"grassheight"},
			legends = {grasslegend},
			curveLabels = {"bla"}
		}
		adding_agents_to_map = Observer{
			subject = herd,
			type = "map",
			observer = map,
			attributes = {"legendstate"}, -- "behaviour"
			legends = {cowlegend_withparasites} -- cowlegend
		}
		table.insert(observers, map)	
		table.insert(observers, adding_agents_to_map)
	end

	-- Charts --

	if (options.observe_grassenergy) then
		energychart = Observer{
			type 		= "chart",
			subject		= pasture,
			attributes	= {"amount_of_grass"},
			curveLabels	= {"Total grass' energy"},
			--attributes	= {"amount_of_cowenergy" , "amount_of_grass"},
			--curveLabels	= {"Total cows' energy" , "Total grass' energy"},
			title		= "Energy",
			xLabel		= "time steps",
			yLabel		= "energy units",
			legends = {legend_chart_grass}
		}
		table.insert(observers, energychart)
	end

	if (options.observe_cownumbers) then
		local cowchart = Observer{
			type 		= "chart",
			subject		= pasture,
			attributes	= {"number_of_cows", "number_of_cows_greedy", "number_of_cows_coop", "zero"},
			curveLabels	= {"Number of cows", "Egoist cows", "Cooperative cows"," "},
			title		= "Number of cows",
			xLabel		= "time steps",
			yLabel		= "number",
			legends = {legend_chart_totalcows, legend_chart_egoist, legend_chart_cooperative, legend_chart_zero}

		}
		table.insert(observers, cowchart)
	end

	if (options.observe_grass_and_cows) then
		local allchart = Observer{
			type 		= "chart",
			subject		= pasture,
			attributes	= {"number_of_cows", "number_of_cows_greedy", "number_of_cows_coop","amount_of_grass_lower", "zero"},
			curveLabels	= {"Number of cows", "Egoist cows", "Cooperative cows", "Grass level"," "},
			title		= "Number of cows",
			xLabel		= "time steps",
			yLabel		= "number",
			legends = {legend_chart_totalcows, legend_chart_egoist, legend_chart_cooperative, legend_chart_grass, legend_chart_zero}
		}
		table.insert(observers, allchart)
	end

	if (options.observe_parasites) and (options.parasites) then
		parasiteschart = Observer{
			type 		= "chart",
			subject 	= pasture,
			attributes	= {"percentage_infected", "percentage_infected_greedy", "percentage_infected_coop", "zero"},
			curveLabels	= {"All cows", "Egoist", "Cooperative", " "},
			title		= "Percentage of infected cows",
			xLabel		= "time steps",
			yLabel		= "number",
			legends = {legend_chart_totalcows, legend_chart_egoist, legend_chart_cooperative, legend_chart_zero}
		}
		table.insert(observers, parasiteschart)
	end

	if (options.observe_averagecows) then
		local averagecowchart = Observer{
			subject = pasture,
			attributes = {"average_energy_coop", "average_energy_greedy"},
			curveLabels = {"average_energy_coop", "average_energy_greedy"},
			--attributes = {"average_energy", "average_energy_coop", "average_energy_greedy"},
			--curveLabels = {"average_cow_energy", "average_energy_coop", "average_energy_greedy"},
			title       	= "Average cow",
			xLabel      	= "days",
			yLabel      	= "energy",
			type ="chart",
			legends = {legend_chart_cooperative, legend_chart_egoist}
		}
		table.insert(observers, averagecowchart)
	end

	----------------
	-- SIMULATION --
	----------------
	t:execute(totaltimesteps)
	print()
	print("There is "..pasture.number_of_cows.." left:\n   "..pasture.number_of_cows_coop.." cooperative and\n   "..pasture.number_of_cows_greedy.." greedy.")



	-- Kill the observers
	if (options.observers) and (options.killobservers) then
		for key,obs in ipairs(observers) do
			obs:kill()
		end
	end

end -- of model function


-- Now run it!
run_model(options, totaltimesteps)
print("Finished!")